package com.wyattjmiller.dottyrevised

import java.util.*

class DotsGame private constructor() {
    enum class AddDotStatus {
        Added, Rejected, Removed
    }

    var movesLeft = 0
        private set
    var score = 0
        private set
    private val mDots: Array<Array<Dot?>>

    // Return list of selected dots
    val selectedDots: ArrayList<Dot?>

    fun getDot(row: Int, col: Int): Dot? {
        return if (row >= GRID_SIZE || row < 0 || col >= GRID_SIZE || col < 0) {
            null
        } else {
            mDots[row][col]
        }
    }

    // Return the last selected dot
    val lastSelectedDot: Dot?
        get() = if (selectedDots.size > 0) {
            selectedDots[selectedDots.size - 1]
        } else {
            null
        }

    // Return the lowest selected dot in each column
    val lowestSelectedDots: ArrayList<Dot?>
        get() {
            val dots = ArrayList<Dot?>()
            for (col in 0 until GRID_SIZE) {
                for (row in GRID_SIZE - 1 downTo 0) {
                    if (mDots[row][col]!!.selected) {
                        dots.add(mDots[row][col])
                        break
                    }
                }
            }
            return dots
        }

    // Clear the list of selected dots
    fun clearSelectedDots() {

        // Reset board so none selected
        for (dot in selectedDots) {
            dot!!.selected = false
        }
        selectedDots.clear()
    }

    // Attempt to add the dot to the list of selected dots
    fun addSelectedDot(dot: Dot): AddDotStatus {
        var status = AddDotStatus.Rejected

        // Check if first dot selected
        if (selectedDots.size == 0) {
            selectedDots.add(dot)
            dot.selected = true
            status = AddDotStatus.Added
        } else {
            if (!dot.selected) {
                // Make sure new is same color and adjacent to last selected dot
                val lastDot = lastSelectedDot
                if (lastDot!!.color == dot.color && lastDot.isAdjacent(dot)) {
                    selectedDots.add(dot)
                    dot.selected = true
                    status = AddDotStatus.Added
                }
            } else if (selectedDots.size > 1) {
                // Dot is already selected, so remove last dot if backtracking
                val secondLast = selectedDots[selectedDots.size - 2]
                if (secondLast == dot) {
                    val removedDot = selectedDots.removeAt(
                        selectedDots.size - 1
                    )
                    removedDot!!.selected = false
                    status = AddDotStatus.Removed
                }
            }
        }
        return status
    }

    // Sort by rows asc
    private fun sortSelectedDots() {
        Collections.sort(selectedDots, object : Comparator<Dot?> {
            override fun compare(dot1: Dot?, dot2: Dot?): Int {
                return dot1!!.row - dot2!!.row
            }
        })
    }

    // Call after completing a dot path to relocate the dots and update the score and moves
    fun finishMove() {
        if (selectedDots.size > 1) {
            // Sort by row so dots are processed top-down
            sortSelectedDots()

            // Move all dots above each selected dot down by changing color
            for (dot in selectedDots) {
                for (row in dot!!.row downTo 1) {
                    val dotCurrent = mDots[row][dot.col]
                    val dotAbove = mDots[row - 1][dot.col]
                    dotCurrent!!.color = dotAbove!!.color
                }

                // Add new dot at top
                val topDot = mDots[0][dot.col]
                topDot!!.setRandomColor()
            }
            score += selectedDots.size
            movesLeft--
            clearSelectedDots()
        }
    }

    // Start a new game
    fun newGame() {
        score = 0
        movesLeft = INIT_MOVES
        for (row in 0 until GRID_SIZE) {
            for (col in 0 until GRID_SIZE) {
                mDots[row][col]!!.setRandomColor()
            }
        }
    }

    // Determine if the game is over
    val isGameOver: Boolean
        get() = movesLeft == 0

    companion object {
        var NUM_COLORS = 5
        var GRID_SIZE = 6
        var INIT_MOVES = 10
        private var mDotsGame: DotsGame? = null
        val instance: DotsGame?
            get() {
                if (mDotsGame == null) {
                    mDotsGame = DotsGame()
                }
                return mDotsGame
            }
    }

    init {

        // Create dots for the 2d array
        mDots = Array(GRID_SIZE) {
            arrayOfNulls(
                GRID_SIZE
            )
        }
        for (row in 0 until GRID_SIZE) {
            for (col in 0 until GRID_SIZE) {
                mDots[row][col] = Dot(row, col)
            }
        }
        selectedDots = ArrayList<Dot?>()
    }
}