package com.wyattjmiller.dottyrevised

import android.animation.Animator
import android.animation.AnimatorListenerAdapter
import android.animation.AnimatorSet
import android.animation.ValueAnimator
import android.content.Context
import android.graphics.Canvas
import android.graphics.Paint
import android.graphics.Path
import android.util.AttributeSet
import android.view.MotionEvent
import android.view.View
import android.view.animation.AccelerateInterpolator
import android.view.animation.BounceInterpolator


class DotsGrid(context: Context?, attrs: AttributeSet?) : View(context, attrs) {
    private val mGame: DotsGame = DotsGame.instance!!
    private val DOT_RADIUS = 40
    private var mAnimatorSet: AnimatorSet = AnimatorSet()

    enum class DotSelectionStatus {
        First, Additional, Last
    }

    interface DotsGridListener {
        fun onDotSelected(dot: Dot?, status: DotSelectionStatus?)
        fun onAnimationFinished()
    }

    private var mGridListener: DotsGridListener? = null
    private val mDotColors: IntArray = resources.getIntArray(R.array.dotColors)
    private var mCellWidth = 0
    private var mCellHeight = 0
    private val mDotPaint: Paint = Paint(Paint.ANTI_ALIAS_FLAG)
    private val mPathPaint: Paint = Paint(Paint.ANTI_ALIAS_FLAG)

    override fun onSizeChanged(width: Int, height: Int, oldWidth: Int, oldHeight: Int) {
        val boardWidth = width - paddingLeft - paddingRight
        val boardHeight = height - paddingTop - paddingBottom
        mCellWidth = boardWidth / DotsGame.GRID_SIZE
        mCellHeight = boardHeight / DotsGame.GRID_SIZE
        resetDots()
    }

    override fun onDraw(canvas: Canvas) {
        super.onDraw(canvas)

        // Draw dots
        for (row in 0 until DotsGame.GRID_SIZE) {
            for (col in 0 until DotsGame.GRID_SIZE) {
                val dot = mGame.getDot(row, col)
                mDotPaint.color = mDotColors[dot!!.color]
                canvas.drawCircle(dot.centerX, dot.centerY, dot.radius, mDotPaint)
            }
        }

        if (!mAnimatorSet.isRunning) {
            // Draw connector
            val selectedDots = mGame.selectedDots
            if (selectedDots.isNotEmpty()) {
                val path = Path()
                var dot = selectedDots[0]
                path.moveTo(dot!!.centerX, dot.centerY)
                for (i in 1 until selectedDots.size) {
                    dot = selectedDots[i]
                    path.lineTo(dot!!.centerX, dot.centerY)
                }
                mPathPaint.color = mDotColors[dot!!.color]
                canvas.drawPath(path, mPathPaint)
            }
        }
    }

    override fun onTouchEvent(event: MotionEvent): Boolean {

        // Only execute when a listener exists
        if (mGridListener == null || mAnimatorSet.isRunning) return true

        // Determine which dot is pressed
        val x = event.x.toInt()
        val y = event.y.toInt()
        val col = x / mCellWidth
        val row = y / mCellHeight
        var selectedDot = mGame.getDot(row, col)

        // Return previously selected dot if touch moves outside the grid
        if (selectedDot == null) {
            selectedDot = mGame.lastSelectedDot
        }

        // Notify activity that a dot is selected
        if (event.action == MotionEvent.ACTION_DOWN) {
            mGridListener!!.onDotSelected(selectedDot, DotSelectionStatus.First)
        } else if (event.action == MotionEvent.ACTION_MOVE) {
            mGridListener!!.onDotSelected(selectedDot, DotSelectionStatus.Additional)
        } else if (event.action == MotionEvent.ACTION_UP) {
            mGridListener!!.onDotSelected(selectedDot, DotSelectionStatus.Last)
        }
        return true
    }

    fun setGridListener(gridListener: DotsGridListener?) {
        mGridListener = gridListener
    }

    fun animateDots() {
        // For storing many animations
        val animations: ArrayList<Animator> = ArrayList()

        val lowestDots = mGame.lowestSelectedDots
        for (dot in lowestDots) {
            var rowsToMove = 1
            for (row in dot!!.row - 1 downTo 0) {
                val dotToMove = mGame.getDot(row, dot.col)
                if (dotToMove!!.selected) {
                    rowsToMove++
                } else {
                    val targetY = dotToMove!!.centerY +
                            rowsToMove * mCellHeight
                    animations.add(
                        getFallingAnimator(
                            dotToMove!!,
                            targetY
                        )!!
                    )
                }
            }
        }


        // Get an animation to make selected dots disappear
        animations.add(getDisappearingAnimator())

        // Play animations (just one right now) together, then reset radius to full size
        mAnimatorSet = AnimatorSet()
        mAnimatorSet.playTogether(animations)
        mAnimatorSet.addListener(object : AnimatorListenerAdapter() {
            override fun onAnimationEnd(animation: Animator) {
                resetDots()
                mGridListener!!.onAnimationFinished()
            }
        })

        mAnimatorSet.start()
    }

    private fun getDisappearingAnimator(): ValueAnimator {
        val animator = ValueAnimator.ofFloat(1f, 0f)
        animator?.duration = 100
        animator?.interpolator = AccelerateInterpolator()
        animator?.addUpdateListener { animation ->
            for (dot in mGame.selectedDots) {
                dot!!.radius = DOT_RADIUS * animation.animatedValue as Float
            }
            invalidate()
        }
        return animator
    }

    private fun getFallingAnimator(dot: Dot, destinationY: Float): ValueAnimator? {
        val animator = ValueAnimator.ofFloat(dot.centerY, destinationY)
        animator?.duration = 300
        animator?.interpolator = BounceInterpolator()
        animator?.addUpdateListener { animation ->
            dot.centerY = animation.animatedValue as Float
            invalidate()
        }
        return animator
    }

    private fun resetDots() {
        for (row in 0 until DotsGame.GRID_SIZE) {
            for (col in 0 until DotsGame.GRID_SIZE) {
                val dot = mGame.getDot(row, col)
                dot!!.radius = DOT_RADIUS.toFloat()
                dot.centerX = (col * mCellWidth + mCellWidth / 2).toFloat()
                dot.centerY = (row * mCellHeight + mCellHeight / 2).toFloat()
            }
        }
    }

    init {
        mPathPaint.strokeWidth = 10f
        mPathPaint.style = Paint.Style.STROKE
    }
}